package My::Test::Module;

use 5.008;

use strict;
use warnings;

use Carp;

our $VERSION = '0.012_02';


1;

__END__

=head1 NAME

My::Test::Module - No functionality, but enough POD to test Test::Pod::LinkCheck::Lite.

=head1 SYNOPSIS

None.

=head1 DESCRIPTION

This module provides no functinoality, and only enough POD to test
C<Test::Pod::LinkCheck::Lite>.

=head1 SUBROUTINES

This module does not implement any of the following subroutines.

=head2 blessed

This subroutine does not exist.

=head1 SUPPORT

This specific module is unspported.

=head1 AUTHOR

Thomas R. Wyant, III F<wyant at cpan dot org>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2021-2024 by Thomas R. Wyant, III

This program is free software; you can redistribute it and/or modify it
under the same terms as Perl 5.10.0. For more details, see the full text
of the licenses in the directory LICENSES.

This program is distributed in the hope that it will be useful, but
without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose.

=cut

# ex: set textwidth=72 :
